import random
from datetime import datetime, timedelta, date
from typing import List
from uuid import uuid4

import pytest
from nothion import NotionClient, PersonalStats
from tickthon import Task, ExpenseLog

from nothion._notion_payloads import NotionPayloads
from nothion._notion_table_headers import ExpensesHeaders


@pytest.fixture(scope="module")
def notion_client(notion_info):
    return NotionClient(notion_info["auth_secret"])


def test_get_active_tasks(notion_client):
    active_tasks = notion_client.get_active_tasks()

    assert len(active_tasks) > 0
    assert isinstance(active_tasks, List) and all(isinstance(i, Task) for i in active_tasks)


def test_get_task_by_etag(notion_client):
    expected_task = Task(ticktick_id="hy76b3d2c8e60f1472064fte",
                         ticktick_etag="9durj438",
                         created_date="9999-09-09",
                         status=2,
                         title="Test Existing Task Static",
                         focus_time=0.9,
                         deleted=0,
                         tags=("test", "existing"),
                         project_id="t542b6d8e9f2de3c5d6e7f8a9s2h",
                         timezone="America/Bogota",
                         due_date="9999-09-09",
                         )

    task = notion_client.get_task_by_etag("9durj438")

    assert task == expected_task


def test_get_task_that_does_not_exist(notion_client):
    task = notion_client.get_task_by_etag("0testdoesntexisttask0")

    assert task is None


def test_get_task_with_missing_properties(notion_client):
    expected_task = Task(ticktick_id="tg81h23oi12h3jkh2720fu321",
                         ticktick_etag="d9iej37s",
                         created_date="2099-09-09",
                         status=2,
                         title="Test Existing Task With Missing Data",
                         )

    task = notion_client.get_task_by_etag("d9iej37s")

    assert task == expected_task


def test_get_notion_id(notion_client):
    expected_notion_id = "f088993635c340cc8e98298ab93ed685"

    notion_id = notion_client.get_task_notion_id("muu17zqq")

    assert notion_id == expected_notion_id


@pytest.mark.parametrize("task_etag, expected_status", [
    # Test with a test task
    ("muu17zqq", True),

    # Test with a task that does not exist
    ("0testtask0", False),
])
def test_is_task_already_created(notion_client, task_etag, expected_status):
    is_task_created = notion_client.is_task_already_created(Task(ticktick_etag=task_etag, created_date="", title="",
                                                                 ticktick_id=""))

    assert is_task_created == expected_status


def test_create_task(notion_client):
    task_id = uuid4().hex
    expected_task = Task(ticktick_id=task_id,
                         ticktick_etag="deletetk",
                         created_date="9999-09-09",
                         status=0,
                         title="Test Task to Delete",
                         focus_time=0.9,
                         tags=("test", "existing", "delete"),
                         project_id="a123a4b5c6d7e8f9a0b1c2d3s4h",
                         timezone="America/Bogota",
                         due_date="9999-09-09",
                         )

    notion_client.create_task(expected_task)

    task = notion_client.get_task_by_etag("deletetk")
    assert task == expected_task

    notion_client.delete_task(expected_task)
    assert notion_client.is_task_already_created(expected_task) is False


def test_complete_task(notion_client):
    task_id = uuid4().hex
    expected_task = Task(ticktick_id=task_id,
                         ticktick_etag="complete",
                         created_date="9999-09-09",
                         status=0,
                         title="Test Task to Complete",
                         focus_time=0.9,
                         tags=("test", "existing", "complete"),
                         project_id="f9ri34b5c6f7rh29a0b1f9eo2ln",
                         timezone="America/Bogota",
                         due_date="9999-09-09",
                         )

    notion_client.create_task(expected_task)
    notion_client.complete_task(expected_task)

    task = notion_client.get_task_by_etag(expected_task.ticktick_etag)
    assert task.status == 2

    notion_client.delete_task(expected_task)
    assert notion_client.is_task_already_created(expected_task) is False


def test_update_task(notion_client):
    expected_task = Task(ticktick_id="a7f9b3d2c8e60f1472065ac4",
                         ticktick_etag="muu17zqq",
                         created_date="9999-09-09",
                         status=2,
                         title="Test Existing Task",
                         focus_time=random.random(),
                         tags=("test", "existing"),
                         project_id="4a72b6d8e9f2103c5d6e7f8a9b0c",
                         timezone="America/Bogota",
                         due_date="9999-09-09",
                         )

    original_task = notion_client.get_task_by_etag("muu17zqq")
    notion_client.update_task(expected_task)
    updated_task = notion_client.get_task_by_etag("muu17zqq")

    assert updated_task == expected_task
    assert updated_task.title == original_task.title
    assert updated_task.focus_time != original_task.focus_time


def test_add_expense_log(notion_client):
    expected_expense_log = ExpenseLog(date="9999-09-09", expense=99.9, product="Test Expense Log")

    expense_log = notion_client.add_expense_log(expected_expense_log)

    expense_log_entry = notion_client.notion_api.get_table_entry(expense_log["id"])
    expense_log_properties = expense_log_entry["properties"]
    assert expense_log_properties[ExpensesHeaders.DATE.value]["date"]["start"] == expected_expense_log.date
    assert expense_log_properties[ExpensesHeaders.EXPENSE.value]["number"] == expected_expense_log.expense
    assert (expense_log_properties[ExpensesHeaders.PRODUCT.value]["title"][0]["text"]["content"]
            == expected_expense_log.product)

    notion_client.notion_api.update_table_entry(expense_log["id"], NotionPayloads.delete_table_entry())


def test_get_incomplete_stats_dates(notion_client):
    stats_date = datetime.now() + timedelta(days=2)

    incomplete_dates = notion_client.get_incomplete_stats_dates(stats_date)

    assert len(incomplete_dates) >= 2
    assert (isinstance(incomplete_dates, List) and
            all(datetime.strptime(i, '%Y-%m-%d') for i in incomplete_dates))


def test_update_stats_row(notion_client):
    notion_api = notion_client.notion_api
    expected_stat = PersonalStats(date="1999-09-09",
                                  weight=0,
                                  work_time=99.9,
                                  leisure_time=99.9,
                                  focus_time=random.random())

    original_stat = notion_client._parse_stats_rows(notion_api.get_table_entry("c568738e82a24b258071e5412db89a2f"))[0]
    notion_client.update_stat(expected_stat)
    updated_stat = notion_client._parse_stats_rows(notion_api.get_table_entry("c568738e82a24b258071e5412db89a2f"))[0]

    assert updated_stat == expected_stat
    assert updated_stat.date == original_stat.date
    assert updated_stat.focus_time != original_stat.focus_time


@pytest.mark.parametrize("start_date, end_date, expected_stats", [
    # Test start date before end date
    (date(2023, 1, 1), date(2023, 1, 3),
     [PersonalStats(date='2023-01-01', work_time=2.03, leisure_time=6.5, focus_time=0, weight=0),
      PersonalStats(date='2023-01-02', work_time=3.24, leisure_time=3.24, focus_time=3.12, weight=0),
      PersonalStats(date='2023-01-03', work_time=7.52, leisure_time=1.52, focus_time=6.33, weight=0)]),

    # Test start date equal to end date
    (date(2023, 1, 1), date(2023, 1, 1),
     [PersonalStats(date='2023-01-01', work_time=2.03, leisure_time=6.5, focus_time=0, weight=0)]),

    # Test start date after end date
    (date(2023, 1, 3), date(2023, 1, 1), []),
])
def test_get_stats_between_dates(notion_client, start_date, end_date, expected_stats):
    stats = notion_client.get_stats_between_dates(start_date, end_date)
    assert stats == expected_stats



import datetime
from typing import List, Optional

from . import PersonalStats
from ._config import NT_TASKS_DB_ID, NT_STATS_DB_ID
from ._notion_payloads import NotionPayloads
from ._notion_table_headers import TasksHeaders, StatsHeaders
from tickthon import Task, ExpenseLog

from ._notion_api import NotionAPI


class NotionClient:

    def __init__(self, auth_secret: str):
        self.notion_api = NotionAPI(auth_secret)
        self.active_tasks: List[Task] = []

    @staticmethod
    def _parse_notion_tasks(raw_tasks: List[dict] | dict) -> List[Task]:
        """Parses the raw tasks from Notion into Task objects."""

        if not isinstance(raw_tasks, list):
            raw_tasks = [raw_tasks]

        parsed_tasks = []
        for raw_task in raw_tasks:
            task_properties = raw_task["properties"]

            timezone = ""
            if task_properties[TasksHeaders.TIMEZONE.value]["rich_text"]:
                timezone = task_properties[TasksHeaders.TIMEZONE.value]["rich_text"][0]["plain_text"]

            due_date = ""
            if task_properties[TasksHeaders.DUE_DATE.value]["date"]:
                due_date = task_properties[TasksHeaders.DUE_DATE.value]["date"]["start"]

            created_date = ""
            if task_properties[TasksHeaders.CREATED_DATE.value]["date"]:
                created_date = task_properties[TasksHeaders.CREATED_DATE.value]["date"]["start"]

            project_id = ""
            if task_properties[TasksHeaders.PROJECT_ID.value]["rich_text"]:
                project_id = task_properties[TasksHeaders.PROJECT_ID.value]["rich_text"][0]["plain_text"]

            parsed_tasks.append(Task(title=task_properties[TasksHeaders.TASK.value]["title"][0]["plain_text"],
                                     status=2 if task_properties[TasksHeaders.DONE.value]["checkbox"] else 0,
                                     ticktick_id=task_properties[TasksHeaders.TICKTICK_ID.value]
                                     ["rich_text"][0]["plain_text"],
                                     ticktick_etag=task_properties[TasksHeaders.TICKTICK_ETAG.value]
                                     ["rich_text"][0]["plain_text"],
                                     created_date=created_date,
                                     focus_time=task_properties[TasksHeaders.FOCUS_TIME.value]["number"],
                                     deleted=int(raw_task["archived"]),
                                     tags=tuple([tag["name"] for tag in task_properties[TasksHeaders.TAGS.value]
                                                                                       ["multi_select"]]),
                                     project_id=project_id,
                                     timezone=timezone,
                                     due_date=due_date))

        return parsed_tasks

    def get_active_tasks(self) -> List[Task]:
        """Gets all active tasks from Notion that are not done."""
        payload = NotionPayloads.get_active_tasks()

        raw_tasks = self.notion_api.query_table(NT_TASKS_DB_ID, payload)
        notion_tasks = self._parse_notion_tasks(raw_tasks)

        self.active_tasks = notion_tasks
        return notion_tasks

    def get_task_by_etag(self, ticktick_etag: str) -> Optional[Task]:
        """Gets the task from Notion that have the given ticktick etag."""
        payload = NotionPayloads.get_task_by_etag(ticktick_etag)
        raw_tasks = self.notion_api.query_table(NT_TASKS_DB_ID, payload)

        notion_tasks = self._parse_notion_tasks(raw_tasks)
        if notion_tasks:
            return notion_tasks[0]
        return None

    def delete_task(self, task: Task):
        """Deletes a task from Notion."""
        task_payload = NotionPayloads.get_task_by_etag(task.ticktick_etag)
        raw_tasks = self.notion_api.query_table(NT_TASKS_DB_ID, task_payload)

        delete_payload = NotionPayloads.delete_table_entry()
        for raw_task in raw_tasks:
            page_id = raw_task["id"]
            self.notion_api.update_table_entry(page_id, delete_payload)

    def get_task_notion_id(self, ticktick_etag: str) -> str:
        """Gets the Notion ID of a task."""
        payload = NotionPayloads.get_task_by_etag(ticktick_etag)
        raw_tasks = self.notion_api.query_table(NT_TASKS_DB_ID, payload)

        return raw_tasks[0]["id"].replace("-", "")

    def is_task_already_created(self, task: Task) -> bool:
        """Checks if a task is already created in Notion."""
        payload = NotionPayloads.get_task_by_etag(task.ticktick_etag)
        raw_tasks = self.notion_api.query_table(NT_TASKS_DB_ID, payload)
        return len(raw_tasks) > 0

    def create_task(self, task: Task) -> Optional[dict]:
        """Creates a task in Notion.

        Args:
            task: The task to create.

        Returns:
            The response from Notion if the task was created.
        """

        payload = NotionPayloads.create_task(task)

        if not self.is_task_already_created(task):
            return self.notion_api.create_table_entry(payload)
        return None

    def update_task(self, task: Task):
        """Updates a task in Notion."""
        page_id = self.get_task_notion_id(task.ticktick_etag)
        payload = NotionPayloads.update_task(task)
        self.notion_api.update_table_entry(page_id, payload)

    def complete_task(self, task: Task):
        page_id = self.get_task_notion_id(task.ticktick_etag)
        payload = NotionPayloads.complete_task()
        self.notion_api.update_table_entry(page_id, payload)

    def add_expense_log(self, expense_log: ExpenseLog) -> dict:
        """Adds an expense log to the expenses DB in Notion."""
        payload = NotionPayloads.create_expense_log(expense_log)
        return self.notion_api.create_table_entry(payload)

    @staticmethod
    def _parse_stats_rows(rows: List[dict] | dict) -> List[PersonalStats]:
        """Parses the raw stats rows from Notion into PersonalStats objects."""
        if not isinstance(rows, List):
            rows = [rows]

        rows_parsed = []
        for row in rows:
            row_properties = row["properties"]
            rows_parsed.append(PersonalStats(date=row_properties[StatsHeaders.DATE.value]["date"]["start"],
                                             weight=row_properties[StatsHeaders.WEIGHT.value]["number"] or 0,
                                             work_time=row_properties[StatsHeaders.WORK_TIME.value]["number"] or 0,
                                             leisure_time=row_properties[StatsHeaders.LEISURE_TIME.value]
                                                                        ["number"] or 0,
                                             focus_time=row_properties[StatsHeaders.FOCUS_TIME.value]["number"] or 0))
        return rows_parsed

    def _get_last_checked(self) -> Optional[PersonalStats]:
        """Gets the last checked row from the stats in Notion database."""
        checked_rows = self.notion_api.query_table(NT_STATS_DB_ID, NotionPayloads.get_checked_stats_rows())
        if checked_rows:
            return self._parse_stats_rows(checked_rows[-1])[0]
        return None

    def get_incomplete_stats_dates(self, limit_date: datetime.date) -> List[str]:
        """Gets the dates that are incomplete in the stats database starting 14 days before the limit date.

        Args:
            limit_date: The limit date that is checked to get the incomplete dates.

        Returns:
            A list of dates in format YYYY-MM-DD.
        """
        initial_date = None
        last_checked_row = self._get_last_checked()
        if last_checked_row:
            current_date = datetime.datetime.strptime(last_checked_row.date, "%Y-%m-%d")
            initial_date = current_date - datetime.timedelta(days=14)

        if initial_date and (limit_date.year > initial_date.year):
            initial_date = None

        payload = NotionPayloads.get_data_between_dates(initial_date, limit_date)
        incomplete_rows = self._parse_stats_rows(self.notion_api.query_table(NT_STATS_DB_ID, payload))

        return [row.date for row in incomplete_rows]

    def update_stat(self, stat_data: PersonalStats):
        """Updates a row in the stats database in Notion."""
        date_row = self.notion_api.query_table(NT_STATS_DB_ID, NotionPayloads.get_date_rows(stat_data.date))
        row_id = date_row[0]["id"]

        notion_row_data_raw = self.notion_api.get_table_entry(row_id)
        notion_row_data = self._parse_stats_rows(notion_row_data_raw)[0]

        if stat_data != notion_row_data:
            self.notion_api.update_table_entry(row_id, NotionPayloads.update_stat(stat_data))

    def get_stats_between_dates(self, start_date: datetime.date, end_date: datetime.date) -> List[PersonalStats]:
        raw_data = self.notion_api.query_table(NT_STATS_DB_ID, NotionPayloads.get_data_between_dates(start_date,
                                                                                                     end_date))
        return self._parse_stats_rows(raw_data)



import json
from typing import List

import requests


class NotionAPI:
    BASE_URL = "https://api.notion.com/v1"
    NOTION_VERSION = "2022-06-28"

    def __init__(self, auth_secret: str):
        self._auth_secret = auth_secret

    def _default_headers(self) -> dict:
        """Returns the headers required for the Notion API requests."""
        return {
            "content-type": "application/json",
            "Notion-Version": self.NOTION_VERSION,
            "Authorization": f"Bearer {self._auth_secret}"
        }

    def query_table(self, table_id: str, query: dict) -> List[dict]:
        next_page_id = None
        first_request = True
        all_results = []

        while next_page_id or first_request:
            first_request = False

            response = requests.post(url=self.BASE_URL + "/databases/" + table_id + "/query",
                                     data=json.dumps(query),
                                     headers=self._default_headers())
            response.raise_for_status()

            all_results += response.json().get("results")
            next_page_id = response.json().get("next_cursor")

            if next_page_id:
                query["start_cursor"] = next_page_id

        return all_results

    def create_table_entry(self, payload: str) -> dict:
        response = requests.post(url=self.BASE_URL + "/pages",
                                 data=payload,
                                 headers=self._default_headers())
        response.raise_for_status()
        return response.json()

    def get_table_entry(self, page_id: str) -> dict:
        response = requests.get(url=self.BASE_URL + "/pages/" + page_id,
                                headers=self._default_headers())
        response.raise_for_status()
        return response.json()

    def update_table_entry(self, page_id: str, payload: str):
        response = requests.patch(url=self.BASE_URL + "/pages/" + page_id,
                                  data=payload,
                                  headers=self._default_headers())
        response.raise_for_status()
